
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "base/polygraph.h"

#include <ctype.h>
#include <strstream.h>
#include <iomanip.h>

#include "xstd/Clock.h"
#include "xstd/gadgets.h"
#include "base/polyLogCats.h"
#include "runtime/Connection.h"
#include "runtime/HostMap.h"
#include "runtime/HttpDate.h"
#include "runtime/StatPhaseMgr.h"
#include "runtime/SharedOpts.h"
#include "runtime/ConnMgr.h"
#include "runtime/ErrorMgr.h"
#include "runtime/LogComment.h"
#include "runtime/polyErrors.h"
#include "runtime/globals.h"
#include "csm/ContentMgr.h"
#include "csm/oid2Url.h"
#include "client/CltOpts.h"
#include "client/Client.h"
#include "client/ServerRep.h"
#include "client/PrivCache.h"
#include "client/cltReqTypes.h"
#include "client/CltXact.h"


/* HTTP header strings (to avoid strlen() calls when creating headers) */

// request line parts
static const String rlpGetHttpOrg = "GET /";
static const String rlpGetHttpPxy = "GET http://";
static const String rlpHttp_10    = " HTTP/1.0\r\n";

// full header fields
static const String hfAccept       = "Accept: */*\r\n";
static const String hfConnAliveOrg = "Connection: keep-alive\r\n";
static const String hfConnCloseOrg = "Connection: close\r\n";
static const String hfConnAlivePxy = "Proxy-Connection: keep-alive\r\n";
static const String hfConnClosePxy = "Proxy-Connection: close\r\n";
static const String hfReload       = "Pragma: no-cache\r\n"
                                     "Cache-Control: no-cache\r\n";

// header field prefixes
static const String hfpHost         = "Host: ";
static const String hfpIMS          = "If-Modified-Since: ";
static const String hfpXFailedOid   = "X-Failed-Oid: ";
static const String hfpXNidShortage = "X-Nid-Shortage: ";
static const String hfpXCltWss      = "X-Clt-Wss: ";
static const String hfpXXact        = "X-Xact: ";


void CltXact::reset() {
	Xaction::reset();
	theOwner = 0;
	theConsumedSize = 0;
	theSrvRep = 0;
	theFailedOid = 0;
	theRepHdr.reset();
	theReqType = rtpBasic;
	thePeerState = peerUnknown;

	theLogCat = lgcCltSide;
}

void CltXact::reqType(int aReqType) {
	Assert(rtpBasic <= aReqType && aReqType < rtpEnd);
	theReqType = aReqType;
}

Agent *CltXact::owner() {
	return theOwner;
}

void CltXact::enqueue() {
	theEnqueTime = TheClock;
}

void CltXact::exec(Client *anOwner, Connection *aConn) {
	Assert(theOidRec.wid);
	Assert(anOwner);
	theOwner = anOwner;

	start(aConn);

	theSrvRep = Client::GetServerRep(theOidRec.hidx);

	makeReq(theConn->theWrBuf);
	theConn->theWr.start(this);
	newState(stSpaceWaiting);
}


void CltXact::finish(int errNo) {
	// get rid of reporting debts
	if (theFailedOid > 0)
		theSrvRep->noteFailedOid(theFailedOid);

	// remember public direct oid if the xaction failed
	if (errNo && theOidRec.wid == theSrvRep->id() &&
		!theOidRec.embedded() && theOidRec.oid > 0)
		theSrvRep->noteFailedOid(theOidRec.oid);

	if (!errNo && staleReply())
		errNo = errStaleHit;

	// cache objects with bodies
	if (!errNo && theRepHdr.theContSize >= 0) {
		if (PrivCache *cache = theOwner->privCache())
			cache->putOid(theOidRec, theRepHdr.calcLmt());
	}

	Xaction::finish(errNo);
	theOwner->noteXactDone(this);
}

void CltXact::noteReadReady(int) {
	theConn->read();

	if (theConn->bad()) {
		finish(-1);
		return;
	}

	if (theState == stHdrWaiting) {
		if (theRepHdr.parse(theConn->theRdBuf.content(), theConn->theRdBuf.contSize())) {

			// dump reply header
			if (TheOpts.theDumpFlags(dumpRep, dumpHdr))
				printMsg(theConn->theRdBuf, theRepHdr.theHdrSize);

			if (const int errNo = interpretHeader()) {
				finish(errNo);
				return;
			}

			consume(theRepHdr.theHdrSize);
			Assert(theState == stBodyWaiting);
			// fall through
		} else
		if (theConn->theRdBuf.full()) { // too big header
			finish(errHugeHdr);
			return;
		} else
		if (theConn->atEof()) { // premature end of headers
			finish(theConn->theRdBuf.contSize() ?
				 errPrematureEoh : errNoHdrClose);
			return;
		}
	}

	if (theState == stBodyWaiting) {
		const Size parsedSize = parseContent();
		consume(parsedSize);
		theConn->theRdBuf.pack();

		if (theConsumedSize == theRepSize) { // got all we need
			if (theOidRec.foreign())
				finish(-1); // see errForeignRep elsewhere
			else
			if (theConn->theRdBuf.contSize() && !theConn->reusable())
				finish(errExtraRepData);
			else
				finish(0);
			return;
		}

		// tag without a closing bracket?
		if (theConn->atEof() &&
			theConsumedSize + theConn->theRdBuf.contSize() >= theRepSize) {
			finish(errOpenTag);
			return;
		}
	}

	if (theConn->atEof()) { // premature eof
		finish(errPrematureEof);
		return;
	}
}

void CltXact::noteWriteReady(int) {
	WrBuf &buf = theConn->theWrBuf;
	Assert(buf.contSize() > 0);
	theConn->write();

	if (theConn->bad()) {
		finish(-1);
		return;
	}

	if (buf.empty()) {
		theConn->theWr.stop(this);
		theConn->theRd.start(this);
		TheFileScanner->setPriority(theConn->sock().fd(), fsupBestEffort);
		newState(stHdrWaiting);
	} else
	if (!theConn->theWr.theReserv)
		theConn->theWr.start(this);
}

void CltXact::makeReq(WrBuf &buf) {
	const char *reqStart = buf.space();
	ostrstream os(buf.space(), buf.spaceSize());

	const NetAddr &orig = theSrvRep->addr();

	/* request line */

	// Method and Request-URI prefix (including first '/' in the path)
	if (theConn->addr() != orig) { // proxied connection
		os << rlpGetHttpPxy << orig.addrA();
		if (orig.port() != 80)
			os << ':' << orig.port();
		os << '/';
	} else {
		// transparent folks get an advantage of shorter request-lines
		os << rlpGetHttpOrg;
	}

	Oid2UrlPath(ObjId(theOidRec), os);
	os << rlpHttp_10; // end-of-request-line

	/*
	 * header fields: it is "good practice" to send general-header
	 * fields first, followed by request-header or response-header
	 * fields, and ending with the entity-header fields.
	 */

	/* general-header fields */

	// indicate persistency (HTTP/1.0 defaults to non-persistent)
	if (theConn->reusable())
		os << (TheCltOpts.thePxyAddr ? hfConnAlivePxy : hfConnAliveOrg);

	/* request-header fields */
	os 
		<< hfAccept
		<< hfpHost << orig.addrA() << ':' << orig.port() << crlf
		;

	if (theReqType == rtpIMS) {
		theOidRec.ims(true);
		static const Time minTime(0,0);
		const PrivCache *cache = theOwner->privCache();
		const Time stamp = cache ? 
			cache->stamp(ObjId(theOidRec)) : Clock::TheStartTime;
		HttpDatePrint(os << hfpIMS, Max(minTime, stamp)) << crlf;
	} else
	if (theReqType == rtpReload) {
		theOidRec.reload(true);
		os << hfReload;
	}

	/* entity-header fields */
	os << hfpXXact << TheGroupId << ' ' << theId 
		<< ' ' << ++TheCount << '/' << theSrvRep->lastRepPos() << crlf;
	if (const int shortage = theSrvRep->nidShortage())
		os << hfpXNidShortage << theSrvRep->pendNidPack() << ':' << shortage << crlf;
	os << hfpXCltWss << theSrvRep->wss() << crlf;

	if (!theOidRec.repeat() && theSrvRep->haveFailedOids()) {
		theFailedOid = theSrvRep->failedOid();
		os << hfpXFailedOid << theFailedOid << crlf;
	}

	os << crlf; // end-of-header

	buf.appended(os.tellp());

	// dump request header
	if (TheOpts.theDumpFlags(dumpReq, dumpHdr))
		printMsg(reqStart, buf.space()-reqStart);
}

void CltXact::consume(Size size) {
	theConn->theRdBuf.consumed(size);
	theConsumedSize += size;
}

int CltXact::interpretHeader() {

	// does reply line make sense?
	if (theRepHdr.theHttpVer < 0 || theRepHdr.theStatus < 0)
		return errHttpRLine;

	// check status code
	theHttpStatus = theRepHdr.theStatus;
	if (theHttpStatus != RepHdr::scOK && theHttpStatus != RepHdr::scNotModified)
		return errHttpStatusCode;

	// check content-length, set RepSize
	if (theRepHdr.expectBody()) {
		if (theRepHdr.theContSize < 0)
			return errNoCLen;
		theRepSize = theRepHdr.theHdrSize + theRepHdr.theContSize;
	} else {
		if (theRepHdr.theContSize >= 0)
			return errUnexpectedCLen;
		theRepSize = theRepHdr.theHdrSize;
	}

	// note: read foreign replies, but
	// do not count foreign replies as misses, hits, etc.!
	if (theRepHdr.foreign()) {
		theOidRec.foreign(true);
		if (ReportError(errForeignRep))
			printMsg(theConn->theRdBuf);
	} else {
		// firstHand here means our request reached the server and
		// server's reply reached us
		const bool firstHand = theId.myMutant(theRepHdr.theXactId);

		theOidRec.cachable(theRepHdr.isCachable);
		theOidRec.hit(!firstHand && theRepHdr.theXactId && !theOidRec.ims());

		if (theOidRec.hit()) {
			if (!theOidRec.offeredHit() && !TheCltOpts.ignoreFalseHits)
				return errFalseHit;

			if (!theOidRec.cachable())
				return errUnchbHit;

			if (theOidRec.reload())
				return errReloadHit;
		}

		if (firstHand) {
			checkDateSync();
			// sync oid accounting on first-hand replies
			theSrvRep->syncSrvInfo(theRepHdr);
			theFailedOid = -1; // no reporting dept
		}

		if (TheCltOpts.printFalseMisses && theOidRec.offeredHit() && !theOidRec.hit()) {
			theOidRec.print(cout << "False-Miss: ") << endl;
			printMsg(theConn->theRdBuf, theRepHdr.theHdrSize);
		}
	}
	
	// note if the server will close the connection
	theConn->lastUse(!theRepHdr.keepAlive());

	newState(stBodyWaiting);
	return 0;
}

bool CltXact::staleReply() const {
	return theRepHdr.theExpires >= 0 && 
		theRepHdr.theExpires < theRepHdr.theDate;
}

void CltXact::checkDateSync() {
	if (theRepHdr.theDate < 0) {
		ReportError(errHttpNoDate);
		return;
	}

	const Time maxGap = Time::Sec(60);
	if (theRepHdr.theDate > TheClock.time() + maxGap ||
		theRepHdr.theDate < TheClock.time() - maxGap) {
		if (ReportError(errSyncDate))
			Comment << theRepHdr.theDate << " - " << TheClock.time() 
				<< " = " << (theRepHdr.theDate - TheClock.time()) << endc;
		return;
	}
}

// returns number of bytes that can be consumed
Size CltXact::parseContent() {
	const IOBuf &buf = theConn->theRdBuf;
	const char *beg = buf.content();
	const char *end = beg + Min(buf.contSize(), theRepSize - theConsumedSize);
	const char *p = beg;

	// do not attempt to parse foreign content
	// do not attempt to parse embedded content
	if (theOidRec.foreign() || theOidRec.embedded())
		return end - beg;

	while (p < end) {
		// find next tag opening
		if (const char *tagO = StrBoundChr(p, '<', end)) {
			p = tagO++;
			// find tag closing
			if (const char *tagC = StrBoundChr(tagO, '>', end)) {
				// found complete tag, parse it
				parseTag(tagO, tagC);
				p = tagC+1;
				// continue
			} else
				break; // wait for more data, leave tag prefix
		} else {
			p = end; // no more tags till the end of a buffer
			break;
		}
	}

	return p - beg;
}

void CltXact::parseTag(const char *tagB, const char *tagE) {
	PolyErrors eno = errNone;

	// find "src=" token within a tag
	for (const char *p = tagB;  p && tagE - p >= 6; p = StrBoundAfterSpace(p, tagE)) {

		if (strncasecmp(p, "src=\"", 5) == 0) {
			const char *url = p + 5;
			if ((p = StrBoundChr(url, '"', tagE))) {
				ReqHdr h;
				if (h.parseUrl(url, p) && h.theOidRec.wid && h.theOidRec.oid > 0 && TheContentMgr.validId(h.theOidRec.tid)) {
					// no support for embedded objects on other servers for now
					if (h.theHost) {
						if (!TheHostMap->find(h.theHost, h.theOidRec.hidx) ||
							h.theOidRec.hidx != theOidRec.hidx)
							eno = errForeignHostName;
					}

					if (!eno) {
						// new oid record needs fixing
						h.theOidRec.hidx = theOidRec.hidx;
						h.theOidRec.repeat(theOidRec.repeat());
						theOwner->noteEmbedded(this, h.theOidRec);
						return;
					}
				} else
					eno = errForeignTag;
			} else
				eno = errMalformedTag;
			break;
		}
	}

	if (eno && !TheCltOpts.ignoreBadContTags &&
		ReportError(eno) && TheOpts.theDumpFlags(dumpErr, dumpAny))
		printMsg(tagB, tagE - tagB);
}

void CltXact::logStats(OLog &ol) const {
	Xaction::logStats(ol);
	ol << theOwner->seqvId() << theReqType;
	theRepHdr.store(ol);
}
